# coding: utf-8
import decimal
from django.utils.translation import ugettext_lazy as _


class Tax(object):
    def __init__(self, name):
        self.name = name
        self._unique_id = None
    
    def __str__(self):
        from django.utils.encoding import smart_str
        return smart_str(unicode(self))
    
    def __unicode__(self):
        return self.name
    
    @property
    def unique_id(self):
        if self._unique_id is None:
            return self._get_unique_id()
        return self._unique_id
    
    def _get_id(self):
        raise NotImplemented()
    
    def amount(self, net):
        raise NotImplemented()
    
    def apply(self, net):
        return net + self.amount(net)
    
    def reverse(self, gross):
        raise NotImplemented()


class NoTax(Tax):
    def __init__(self):
        self.name = _('No Tax')
        self._unique_id = 'none'
    
    def amount(self, net):
        return decimal.Decimal('0')
    
    def reverse(self, gross):
        return gross
NO_TAX = NoTax()


class LinearTax(Tax):
    def __init__(self, name, percent):
        super(LinearTax, self).__init__(name)
        if not isinstance(percent, decimal.Decimal):
            percent = decimal.Decimal(str(percent) or 'NaN')
        self.percent = percent
    
    def _get_unique_id(self):
        return 'linear-%s' % str(self.percent)
    
    def amount(self, net):
        return net * self.percent
    
    def reverse(self, gross, currency=None):
        return gross / (1 + self.percent)


class MultiTax(Tax):
    def __init__(self, taxes, name=None):
        if name is None:
            super(MultiTax, self).__init__(' + '.join([t.name for t in taxes]))
        else:
            super(MultiTax, self).__init__(name)
        self.taxes = taxes
    
    def _get_unique_id(self):
        return 'multi-%s' % '-'.join(list(sort([t.get_id() for t in self.taxes])))
    
    def amount(self, net):
        amount = decimal.Decimal(0)
        for tax in self.taxes:
            amount += tax.amount(net)
        return amount
    
    def reverse(self, gross, currency=None):
        result = gross
        for tax in reversed(self.taxes):
            result = tax.reverse(result)
        return result






import decimal

def price_amount(value, currency=None):
    from .currency import Currency
    decimal_places = 3
    rounding = decimal.ROUND_HALF_UP
    if not currency is None:
        if not isinstance(currency, Currency):
            currency = Currency(currency)
        if not currency.decimal_places is None:
            decimal_places = currency.decimal_places
        if not currency.rounding is None:
            rounding = currency.rounding
    if not isinstance(value, (decimal.Decimal)):
        value = decimal.Decimal(str(value))
    quantize_str = '0'
    if decimal_places > 0:
        quantize_str = '0.%s' % ('0' * decimal_places)
    return value.quantize(decimal.Decimal(quantize_str), rounding=rounding)




# coding: utf-8

# see http://en.wikipedia.org/wiki/ISO_4217#Active_codes
CURRENCIES = [
    # ISO    ISO  dec  round name                        symbol
    # Code   Num  plac ing
    #             es
    ('AED', '784', 2, None, 'United Arab Emirates dirham', None),
    ('AFN', '971', 2, None, 'Afghan afghani', None),
    ('ALL', '008', 2, None, 'Albanian lek', None),
    ('AMD', '051', 2, None, 'Armenian dram', None),
    ('ANG', '532', 2, None, 'Netherlands Antillean guilder', None),
    ('AOA', '973', 2, None, 'Angolan kwanza', None),
    ('ARS', '032', 2, None, 'Argentine peso', None),
    ('AUD', '036', 2, None, 'Australian dollar', None),
    ('AWG', '533', 2, None, 'Aruban florin', None),
    ('AZN', '944', 2, None, 'Azerbaijani manat', None),
    ('BAM', '977', 2, None, 'Bosnia and Herzegovina convertible mark', None),
    ('BBD', '052', 2, None, 'Barbados dollar', None),
    ('BDT', '050', 2, None, 'Bangladeshi taka', None),
    ('BGN', '975', 2, None, 'Bulgarian lev', None),
    ('BHD', '048', 3, None, 'Bahraini dinar', None),
    ('BIF', '108', 0, None, 'Burundian franc', None),
    ('BMD', '060', 2, None, 'Bermudian dollar (customarily known as Bermuda dollar)', None),
    ('BND', '096', 2, None, 'Brunei dollar', None),
    ('BOB', '068', 2, None, 'Boliviano', None),
    ('BOV', '984', 2, None, 'Bolivian Mvdol (funds code)', None),
    ('BRL', '986', 2, None, 'Brazilian real', None),
    ('BSD', '044', 2, None, 'Bahamian dollar', None),
    ('BTN', '064', 2, None, 'Bhutanese ngultrum', None),
    ('BWP', '072', 2, None, 'Botswana pula', None),
    ('BYR', '974', 0, None, 'Belarusian ruble', None),
    ('BZD', '084', 2, None, 'Belize dollar', None),
    ('CAD', '124', 2, None, 'Canadian dollar', None),
    ('CDF', '976', 2, None, 'Congolese franc', None),
    ('CHE', '947', 2, None, 'WIR Euro (complementary currency)', None),
    ('CHF', '756', 2, None, 'Swiss franc', None),
    ('CHW', '948', 2, None, 'WIR Franc (complementary currency)', None),
    ('CLF', '990', 0, None, 'Unidad de Fomento (funds code)', None),
    ('CLP', '152', 0, None, 'Chilean peso', None),
    ('CNY', '156', 2, None, 'Chinese yuan', None),
    ('COP', '170', 2, None, 'Colombian peso', None),
    ('COU', '970', 2, None, 'Unidad de Valor Real', None),
    ('CRC', '188', 2, None, 'Costa Rican colon', None),
    ('CUC', '931', 2, None, 'Cuban convertible peso', None),
    ('CUP', '192', 2, None, 'Cuban peso', None),
    ('CVE', '132', 0, None, 'Cape Verde escudo', None),
    ('CZK', '203', 2, None, 'Czech koruna', None),
    ('DJF', '262', 0, None, 'Djiboutian franc', None),
    ('DKK', '208', 2, None, 'Danish krone', None),
    ('DOP', '214', 2, None, 'Dominican peso', None),
    ('DZD', '012', 2, None, 'Algerian dinar', None),
    ('EGP', '818', 2, None, 'Egyptian pound', None),
    ('ERN', '232', 2, None, 'Eritrean nakfa', None),
    ('ETB', '230', 2, None, 'Ethiopian birr', None),
    ('EUR', '978', 2, None, 'Euro', u'€'),
    ('FJD', '242', 2, None, 'Fiji dollar', None),
    ('FKP', '238', 2, None, 'Falkland Islands pound', None),
    ('GBP', '826', 2, None, 'Pound sterling', None),
    ('GEL', '981', 2, None, 'Georgian lari', None),
    ('GHS', '936', 2, None, 'Ghanaian cedi', None),
    ('GIP', '292', 2, None, 'Gibraltar pound', None),
    ('GMD', '270', 2, None, 'Gambian dalasi', None),
    ('GNF', '324', 0, None, 'Guinean franc', None),
    ('GTQ', '320', 2, None, 'Guatemalan quetzal', None),
    ('GYD', '328', 2, None, 'Guyanese dollar', None),
    ('HKD', '344', 2, None, 'Hong Kong dollar', None),
    ('HNL', '340', 2, None, 'Honduran lempira', None),
    ('HRK', '191', 2, None, 'Croatian kuna', None),
    ('HTG', '332', 2, None, 'Haitian gourde', None),
    ('HUF', '348', 2, None, 'Hungarian forint', None),
    ('IDR', '360', 0, None, 'Indonesian rupiah', None),
    ('ILS', '376', 2, None, 'Israeli new sheqel', None),
    ('INR', '356', 2, None, 'Indian rupee', None),
    ('IQD', '368', 0, None, 'Iraqi dinar', None),
    ('IRR', '364', 0, None, 'Iranian rial', None),
    ('ISK', '352', 0, None, 'Icelandic króna', None),
    ('JMD', '388', 2, None, 'Jamaican dollar', None),
    ('JOD', '400', 3, None, 'Jordanian dinar', None),
    ('JPY', '392', 0, None, 'Japanese yen', None),
    ('KES', '404', 2, None, 'Kenyan shilling', None),
    ('KGS', '417', 2, None, 'Kyrgyzstani som', None),
    ('KHR', '116', 2, None, 'Cambodian riel', None),
    ('KMF', '174', 0, None, 'Comoro franc', None),
    ('KPW', '408', 0, None, 'North Korean won', None),
    ('KRW', '410', 0, None, 'South Korean won', None),
    ('KWD', '414', 3, None, 'Kuwaiti dinar', None),
    ('KYD', '136', 2, None, 'Cayman Islands dollar', None),
    ('KZT', '398', 2, None, 'Kazakhstani tenge', None),
    ('LAK', '418', 0, None, 'Lao kip', None),
    ('LBP', '422', 0, None, 'Lebanese pound', None),
    ('LKR', '144', 2, None, 'Sri Lankan rupee', None),
    ('LRD', '430', 2, None, 'Liberian dollar', None),
    ('LSL', '426', 2, None, 'Lesotho loti', None),
    ('LTL', '440', 2, None, 'Lithuanian litas', None),
    ('LVL', '428', 2, None, 'Latvian lats', None),
    ('LYD', '434', 3, None, 'Libyan dinar', None),
    ('MAD', '504', 2, None, 'Moroccan dirham', None),
    ('MDL', '498', 2, None, 'Moldovan leu', None),
    #('MGA', '969', 0.7, None, 'Malagasy ariary', None),
    ('MKD', '807', 2, None, 'Macedonian denar', None),
    ('MMK', '104', 0, None, 'Myanma kyat', None),
    ('MNT', '496', 2, None, 'Mongolian tugrik', None),
    ('MOP', '446', 2, None, 'Macanese pataca', None),
    #('MRO', '478', 0.7, None, 'Mauritanian ouguiya', None),
    ('MUR', '480', 2, None, 'Mauritian rupee', None),
    ('MVR', '462', 2, None, 'Maldivian rufiyaa', None),
    ('MWK', '454', 2, None, 'Malawian kwacha', None),
    ('MXN', '484', 2, None, 'Mexican peso', None),
    ('MXV', '979', 2, None, 'Mexican Unidad de Inversion (UDI) (funds code)', None),
    ('MYR', '458', 2, None, 'Malaysian ringgit', None),
    ('MZN', '943', 2, None, 'Mozambican metical', None),
    ('NAD', '516', 2, None, 'Namibian dollar', None),
    ('NGN', '566', 2, None, 'Nigerian naira', None),
    ('NIO', '558', 2, None, 'Nicaraguan córdoba', None),
    ('NOK', '578', 2, None, 'Norwegian krone', None),
    ('NPR', '524', 2, None, 'Nepalese rupee', None),
    ('NZD', '554', 2, None, 'New Zealand dollar', None),
    ('OMR', '512', 3, None, 'Omani rial', None),
    ('PAB', '590', 2, None, 'Panamanian balboa', None),
    ('PEN', '604', 2, None, 'Peruvian nuevo sol', None),
    ('PGK', '598', 2, None, 'Papua New Guinean kina', None),
    ('PHP', '608', 2, None, 'Philippine peso', None),
    ('PKR', '586', 2, None, 'Pakistani rupee', None),
    ('PLN', '985', 2, None, 'Polish złoty', None),
    ('PYG', '600', 0, None, 'Paraguayan guaraní', None),
    ('QAR', '634', 2, None, 'Qatari rial', None),
    ('RON', '946', 2, None, 'Romanian new leu', None),
    ('RSD', '941', 2, None, 'Serbian dinar', None),
    ('RUB', '643', 2, None, 'Russian rouble', None),
    ('RWF', '646', 0, None, 'Rwandan franc', None),
    ('SAR', '682', 2, None, 'Saudi riyal', None),
    ('SBD', '090', 2, None, 'Solomon Islands dollar', None),
    ('SCR', '690', 2, None, 'Seychelles rupee', None),
    ('SDG', '938', 2, None, 'Sudanese pound', None),
    ('SEK', '752', 2, None, 'Swedish krona/kronor', None),
    ('SGD', '702', 2, None, 'Singapore dollar', None),
    ('SHP', '654', 2, None, 'Saint Helena pound', None),
    ('SLL', '694', 0, None, 'Sierra Leonean leone', None),
    ('SOS', '706', 2, None, 'Somali shilling', None),
    ('SRD', '968', 2, None, 'Surinamese dollar', None),
    ('SSP', '728', 2, None, 'South Sudanese pound', None),
    ('STD', '678', 0, None, 'São Tomé and Príncipe dobra', None),
    ('SYP', '760', 2, None, 'Syrian pound', None),
    ('SZL', '748', 2, None, 'Swazi lilangeni', None),
    ('THB', '764', 2, None, 'Thai baht', None),
    ('TJS', '972', 2, None, 'Tajikistani somoni', None),
    ('TMT', '934', 2, None, 'Turkmenistani manat', None),
    ('TND', '788', 3, None, 'Tunisian dinar', None),
    ('TOP', '776', 2, None, 'Tongan paʻanga', None),
    ('TRY', '949', 2, None, 'Turkish lira', 'TL'),
    ('TTD', '780', 2, None, 'Trinidad and Tobago dollar', None),
    ('TWD', '901', 2, None, 'New Taiwan dollar', None),
    ('TZS', '834', 2, None, 'Tanzanian shilling', None),
    ('UAH', '980', 2, None, 'Ukrainian hryvnia', None),
    ('UGX', '800', 2, None, 'Ugandan shilling', None),
    ('USD', '840', 2, None, 'United States dollar', '$'),
    #('USN', '997', 2, None, 'United States dollar (next day) (funds code)', None),
    #('USS', '998', 2, None, 'United States dollar (same day) (funds code) (one source[who?] claims it is no longer used, but it is still on the ISO 4217-MA list)', None),
    ('UYI', '940', 0, None, 'Uruguay Peso en Unidades Indexadas (URUIURUI) (funds code)', None),
    ('UYU', '858', 2, None, 'Uruguayan peso', None),
    ('UZS', '860', 2, None, 'Uzbekistan som', None),
    ('VEF', '937', 2, None, 'Venezuelan bolívar fuerte', None),
    ('VND', '704', 0, None, 'Vietnamese đồng', None),
    ('VUV', '548', 0, None, 'Vanuatu vatu', None),
    ('WST', '882', 2, None, 'Samoan tala', None),
    ('XAF', '950', 0, None, 'CFA franc BEAC', None),
    #('XAG', '961', None, None, 'Silver (one troy ounce)', None),
    #('XAU', '959', None, None, 'Gold (one troy ounce)', None),
    #('XBA', '955', None, None, 'European Composite Unit (EURCO) (bond market unit)', None),
    #('XBB', '956', None, None, 'European Monetary Unit (E.M.U.-6) (bond market unit)', None),
    #('XBC', '957', None, None, 'European Unit of Account 9 (E.U.A.-9) (bond market unit)', None),
    #('XBD', '958', None, None, 'European Unit of Account 17 (E.U.A.-17) (bond market unit)', None),
    ('XCD', '951', 2, None, 'East Caribbean dollar', None),
    #('XDR', '960', None, None, 'Special Drawing Rights', None),
    #('XFU', 'Nil', None, None, 'UIC franc (special settlement currency)', None),
    ('XOF', '952', 0, None, 'CFA Franc BCEAO', None),
    #('XPD', '964', None, None, 'Palladium (one troy ounce)', None),
    ('XPF', '953', 0, None, 'CFP franc', None),
    #('XPT', '962', None, None, 'Platinum (one troy ounce)', None),
    #('XTS', '963', None, None, 'Code reserved for testing purposes', None),
    #('XXX', '999', None, None, 'No currency', None),
    ('YER', '886', 2, None, 'Yemeni rial', None),
    ('ZAR', '710', 2, None, 'South African rand', None),
    ('ZMK', '894', 2, None, 'Zambian kwacha', None),
    ('ZWL', '932', 2, None, 'Zimbabwe dollar'),
]



from django.db import models
from django_deferred_polymorph.models import DeferredPolymorphManager

class TaxManager(DeferredPolymorphManager):
    _cache = {}
    
    def get_for_id(self, id):
        try:
            tax = self.__class__._cache[self.db][id]
        except KeyError:
            # This could raise a DoesNotExist; that's correct behavior and will
            # make sure that only correct ctypes get stored in the cache dict.
            tax = self.get(pk=id).get_real_instance()
            self._add_to_cache(self.db, tax)
        return tax

    def clear_cache(self):
        self.__class__._cache.clear()

    def _add_to_cache(self, using, tax):
        self.__class__._cache.setdefault(using, {})[tax.id] = tax




# coding: utf-8
import decimal
from . import settings as price_settings
from .utils import price_amount
from .currency import Currency
from .tax import NO_TAX


class Price(object):
    def __init__(self, net, currency=None, tax=None, gross=None):
        if currency is None:
            currency = price_settings.DEFAULT_CURRENCY
        if not isinstance(currency, Currency):
            currency = Currency(currency)
        self.currency = currency
        
        if not isinstance(net, decimal.Decimal):
            net = decimal.Decimal(str(net) or 'NaN')
        
        # support tax models
        if tax is not None and hasattr(tax, 'get_tax'):
            tax = tax.get_tax()
        
        # calculate tax, gross
        self._applied_taxes = {}
        if not tax is None and not gross is None:
            # we need to trust the external calculation here
            if not isinstance(gross, decimal.Decimal):
                gross = decimal.Decimal(str(gross or '') or 'NaN')
        elif not tax is None:
            # self.net is still not rounded here, so tax_amount is super-precise ;-)
            gross = tax.apply(net)
        elif not gross is None:
            raise RuntimeError('cannot specify gross amount without tax')
        else:
            # no tax applied
            gross = net
            tax = NO_TAX
        self._applied_taxes[tax.unique_id] = (tax, net, gross)
        self._recalculate_overall()
    
    def _recalculate_overall(self):
        # we pass net/gross through price_amount as late as possible, to avoid
        # removing decimal_places we might need to calculate the right
        # gross or tax. self._applied_taxes always stores the raw values without
        # any rounding. This way we do not loose precision on calculation.
        net = decimal.Decimal('0')
        gross = decimal.Decimal('0')
        for tax, tax_net, tax_gross in self._applied_taxes.values():
            # we have to round every net/gross on its own, otherwise
            # we would risk rounding issues (0.3 + 0.3 = 0.6, rounded
            # 0 + 0 = 1)
            net += price_amount(tax_net, self.currency)
            gross += price_amount(tax_gross, self.currency)
        self.net = net
        self.gross = gross
    
    def __str__(self):
        from django.utils.encoding import smart_str
        return smart_str(unicode(self))
    
    def __unicode__(self):
        from django.utils.translation import ugettext
        return ugettext('%(currency)s %(amount)s') % {
            'amount': self.formatted_gross,
            'currency': self.formatted_currency,
        }
    
    def copy(self):
        from copy import copy
        result = copy(self)
        result._applied_taxes = self._applied_taxes.copy()
        return result
    
    def rounded(self):
        applied_taxes = {}
        for tax, net, gross in self._applied_taxes.values():
            applied_taxes[tax.unique_id] = (
                tax,
                price_amount(net, self.currency),
                price_amount(gross, self.currency),
            )
        return CalculatedPrice(applied_taxes, self.currency)
    
    @property
    def precise_net(self):
        return sum([t[1] for t in self._applied_taxes.values()])
    
    @property
    def precise_gross(self):
        return sum([t[2] for t in self._applied_taxes.values()])
    
    @property
    def precise_tax(self):
        return sum([t[2] - t[1] for t in self._applied_taxes.values()])
    
    @property
    def tax(self):
        return self.gross - self.net
    
    @property
    def applied_tax(self):
        if len(self._applied_taxes) != 1:
            raise RuntimeError('This Price has multiple taxes, use obj.taxes instead')
        return self._applied_taxes.values()[0][0]
    
    @property
    def applied_taxes(self):
        return [
            Price(
                net = net,
                tax = tax,
                gross = gross,
                currency = self.currency,
            )
            for tax, net, gross
            in self._applied_taxes.values()
        ]
    
    @property
    def formatted_currency(self):
        return self.currency.symbol if self.currency.symbol else self.currency.iso_code
    
    def _format_amount(self, value):
        from django.utils.formats import number_format
        # workaround for django not treating decimal_places==0 is implied
        # as prices always are rounded to their decimal_places, see
        # utils.price_amount
        # see https://code.djangoproject.com/ticket/13810
        return number_format(value, self.currency.decimal_places or 0)
    
    @property
    def formatted_net(self):
        return self._format_amount(self.net)
    
    @property
    def formatted_gross(self):
        return self._format_amount(self.gross)
    
    @property
    def formatted_tax(self):
        return self._format_amount(self.tax)
    
    def __add__(self, other):
        # EmptyPrice should work regardless of currency, does not change anything
        if isinstance(other, EmptyPrice):
            self.copy()
        if not isinstance(other, Price):
            raise TypeError('cannot add %s' % type(other))
        if self.currency != other.currency:
            raise TypeError('cannot add different currencies')
        applied_taxes = self._applied_taxes.copy()
        for tax, net, gross in other._applied_taxes.values():
            if tax.unique_id in applied_taxes:
                applied_taxes[tax.unique_id] = (
                    applied_taxes[tax.unique_id][0],
                    applied_taxes[tax.unique_id][1] + net,
                    applied_taxes[tax.unique_id][2] + gross,
                )
            else:
                applied_taxes[tax.unique_id] = (tax, net, gross)
        # filter out NO_TAX, if it is not relevant
        if NO_TAX.unique_id in applied_taxes \
            and applied_taxes[NO_TAX.unique_id][1] == 0 \
            and applied_taxes[NO_TAX.unique_id][2] == 0:
                del applied_taxes[NO_TAX.unique_id]
        return CalculatedPrice(applied_taxes, self.currency)
    
    def __neg__(self):
        applied_taxes = {}
        for tax, net, gross in self._applied_taxes.values():
            applied_taxes[tax.unique_id] = (tax, -net, -gross)
        return CalculatedPrice(applied_taxes, self.currency)
    
    def __mul__(self, factor):
        if not isinstance(factor, (int, long, float, decimal.Decimal)):
            raise TypeError("Cannot multiply with %s" % type(factor))
        if not isinstance(factor, decimal.Decimal):
            factor = decimal.Decimal(str(factor))
        if factor.is_nan():
            raise TypeError("Factor must be a number (!= 'NaN')")
        applied_taxes = {}
        for tax, net, gross in self._applied_taxes.values():
            calc_net = net * factor
            calc_gross = gross * factor
            applied_taxes[tax.unique_id] = (tax, calc_net, calc_gross)
        return CalculatedPrice(applied_taxes, self.currency)
    
    def __div__(self, factor):
        if not isinstance(factor, (int, long, float, decimal.Decimal)):
            raise TypeError("Cannot multiply with %s" % type(factor))
        if not isinstance(factor, decimal.Decimal):
            factor = decimal.Decimal(str(factor))
        if factor.is_nan():
            raise TypeError("Factor must be a number (!= 'NaN')")
        applied_taxes = {}
        for tax, net, gross in self._applied_taxes.values():
            calc_net = net / factor
            calc_gross = gross / factor
            applied_taxes[tax.unique_id] = (tax, calc_net, calc_gross)
        return CalculatedPrice(applied_taxes, self.currency)
    __truediv__ = __div__
    
    # django_ajax hook
    def ajax_data(self):
        return {
            'tax': self.formatted_tax,
            'net': self.formatted_net,
            'gross': self.formatted_gross,
            'currency': self.currency.ajax_data(),
        }


class CalculatedPrice(Price):
    def __init__(self, applied_taxes, currency=None):
        if currency is None:
            currency = price_settings.DEFAULT_CURRENCY
        if not isinstance(currency, Currency):
            currency = Currency(currency)
        self.currency = currency
        
        self._applied_taxes = applied_taxes
        self._recalculate_overall()


class EmptyPrice(Price):
    def __init__(self):
        self.net = decimal.Decimal('0')
        self.currency = Currency(price_settings.DEFAULT_CURRENCY)
        self.gross = decimal.Decimal('0')
        self._applied_taxes = {
            NO_TAX.unique_id: (NO_TAX, self.net, self.gross)
        }
    
    def copy(self):
        return self
    
    def __add__(self, other):
        return other.copy()
    
    def __mul__(self, factor):
        return self
    
    def __div__(self, factor):
        return self
    __truediv__ = __div__




from django.db import models
from . import settings as price_settings
from .price import Price
from .tax import Tax
from .currency import Currency


def south_field_triple(self):
    from south.modelsinspector import introspector
    field_class = self.__class__.__module__ + "." + self.__class__.__name__
    args, kwargs = introspector(self)
    return (field_class, args, kwargs)


# price


class PriceField(models.DecimalField):
    ''' Price '''
    
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_digits', 15)
        kwargs.setdefault('decimal_places', 3)
        super(PriceField, self).__init__(*args, **kwargs)
    
    south_field_triple = south_field_triple


class PrecisePriceField(PriceField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_digits', 21)
        kwargs.setdefault('decimal_places', 9)
        super(PrecisePriceField, self).__init__(*args, **kwargs)


class PriceObjectAttr(object):
    def __init__(self, model, access, name):
        self.model = model
        self.access = access
        self.name = name
    
    def __get__(self, obj, type=None):
        value = getattr(obj, self.access.price, None)
        if value is None:
            return None
        return self.access._get_price(obj, value)
    
    def __set__(self, obj, value):
        if isinstance(value, Price):
            self.access._set_currency(obj, value)
            setattr(obj, self.access.price, self.access._get_price_value(value))
        else:
            setattr(obj, self.access.price, value)


class PriceAccessBase(object):
    def __init__(self, price, tax, currency=None):
        self.price = price
        self.tax = tax
        self.currency = currency
    
    def _get_currency(self, instance):
        if self.currency is None:
            return None # Price() will default to DEFAULT_CURRENCY
        currency = self.currency
        if hasattr(currency, '__call__'):
            currency = currency(instance)
        else:
            if self.currency[0] == '=':
                return Currency(self.currency[1:])
            return getattr(instance, self.currency)
    
    def _set_currency(self, instance, price):
        currency = price.currency
        if not isinstance(currency, Currency):
            currency = Currency(currency)
        if hasattr(self.currency, '__call__'):
            self.currency(instance, currency)
        else:
            if self.currency is None: # we always use the default currency, so make sure it fits
                if currency.iso_code != price_settings.DEFAULT_CURRENCY:
                    raise RuntimeError('cannot change fixed (default) currency')
                return
            elif self.currency[0] == '=': # currency is fixed, make sure it fits
                if currency.iso_code != self.currency[1:]:
                    raise RuntimeError('cannot change fixed currency')
                return
            setattr(instance, self.currency, currency)
    
    def _get_tax(self, instance):
        from django.db.models.fields import FieldDoesNotExist
        from .models import Tax
        tax = self.tax
        if hasattr(self.tax, '__call__'):
            tax = tax(instance)
        if isinstance(tax, Tax):
            return tax
        try:
            field = instance._meta.get_field(tax)
        except FieldDoesNotExist:
            return getattr(instance, tax)
        tax_id = getattr(instance, field.attname)
        return Tax.objects.get_for_id(tax_id)
    
    def _get_price(self, instance, price):
        raise NotImplemented()
    
    def _get_price_value(self, price):
        raise NotImplemented()
    
    def contribute_to_class(self, cls, name):
        setattr(cls, name, PriceObjectAttr(cls, self, name))
        def get_price(s):
            return getattr(s, name)
        setattr(cls, "get_%s" % name, get_price)
        def set_price(s, v):
            return setattr(s, name, v)
        setattr(cls, "set_%s" % name, set_price)


class NetPriceAccess(PriceAccessBase):
    def _get_price(self, instance, price):
        tax = self._get_tax(instance)
        price = Price(
            net = price,
            currency = self._get_currency(instance),
            tax = tax,
        )
        return price
    
    def _get_price_value(self, price):
        return price.net


class GrossPriceAccess(PriceAccessBase):
    def _get_price(self, instance, price):
        tax = self._get_tax(instance)
        net = tax.reverse(price)
        price = Price(
            net = net,
            currency = self._get_currency(instance),
            tax = tax,
            gross = price,
        )
        return price
    
    def _get_price_value(self, price):
        return price.gross


# currency


class CurrencyField(models.CharField):
    def __init__(self, *args, **kwargs):
        from .currencies import CURRENCIES
        kwargs.setdefault('max_length', 3)
        kwargs.setdefault('default', price_settings.DEFAULT_CURRENCY)
        kwargs.setdefault('choices', [(i[0], i[0]) for i in CURRENCIES])
        super(CurrencyField, self).__init__(*args, **kwargs)
    
    south_field_triple = south_field_triple


class CurrencyObjectAttr(object):
    def __init__(self, model, access, name):
        self.model = model
        self.access = access
        self.name = name
    
    def __get__(self, obj, type=None):
        value = getattr(obj, self.access.currency, None)
        if value is None:
            return None
        currency = Currency(
            iso_code = value,
        )
        return currency
    
    def __set__(self, obj, value):
        if isinstance(value, Currency):
            setattr(obj, self.access.currency, self.access._get_currency_value(value))
        else:
            setattr(obj, self.access.currency, value)


class CurrencyAccess(models.CharField):
    def __init__(self, currency):
        self.currency = currency
    
    def _get_currency_value(self, currency):
        return currency.iso_code
    
    def contribute_to_class(self, cls, name):
        setattr(cls, name, CurrencyObjectAttr(cls, self, name))
        def get_currency(s):
            return getattr(s, name)
        setattr(cls, "get_%s" % name, get_currency)
        def set_currency(s, v):
            return setattr(s, name, v)
        setattr(cls, "set_%s" % name, set_currency)


# tax


class TaxField(models.ForeignKey):
    def __init__(self, *args, **kwargs):
        from .models import Tax
        kwargs.setdefault('to', Tax)
        kwargs.setdefault('related_name', '+')
        super(TaxField, self).__init__(*args, **kwargs)
    
    south_field_triple = south_field_triple




from django.contrib import admin
from .models import *


# we do not add admins, this has to be done inside the project
#admin.site.register(LinearTax, admin.ModelAdmin)
#admin.site.register(MultiTax, admin.ModelAdmin)




from setuptools import setup, find_packages

setup(
    name = "django_price",
    version = "0.1.0",
    description = 'Handling price information in django',
    author = 'David Danier',
    author_email = 'david.danier@team23.de',
    url = 'https://github.com/ddanier/django_price',
    long_description=open('README.rst', 'r').read(),
    packages = [
        'django_price',
    ],
    requires = [
        'django(>=1.4)',
        'django_deferred_polymorphic(>=0.1.0)',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Utilities'
    ],
)




# coding: utf-8
from .currencies import CURRENCIES


class Currency(object):
    CURRENCIES = dict([(c[0], c[1:]) for c in CURRENCIES])
    
    def __init__(self, iso_code):
        if not iso_code in self.CURRENCIES:
            raise TypeError('unknown currency (%s)' % iso_code)
        self.iso_code = iso_code
        self.iso_num, self.decimal_places, self.rounding, self.name, self.symbol = self.CURRENCIES[iso_code]
    
    def __str__(self):
        from django.utils.encoding import smart_str
        return smart_str(unicode(self))
    
    def __unicode__(self):
        return self.iso_code
    
    def __eq__(self, other):
        if not isinstance(other, Currency):
            # allow direct comparision to iso codes
            if other in self.CURRENCIES:
                return self.iso_code == other
            return False
        return self.iso_code == other.iso_code
    
    def __ne__(self, other):
        return not self == other
    
    # django_ajax hook
    def ajax_data(self):
        return {
            'iso_code': self.iso_code,
            'name': self.name,
            'symbol': self.symbol,
        }





# coding: utf-8
from django.utils.translation import ugettext_lazy as _
from django.db import models
from django_deferred_polymorph.models import SubDeferredPolymorphBaseModel
import decimal
import datetime
from .manager import TaxManager


# TODO: Versionized Tax (Tax should NEVER get changed, as this may
# create an invalid state if you store net + gross for invoices
class Tax(SubDeferredPolymorphBaseModel):
    name = models.CharField(max_length=25)
    
    created = models.DateTimeField(editable=False, default=datetime.datetime.now)
    modified = models.DateTimeField(editable=False, default=datetime.datetime.now, auto_now=True)
    
    objects = TaxManager()
    
    def __unicode__(self):
        return self.name
    
    @property
    def unique_id(self):
        return self.get_tax().unique_id
    
    def amount(self, net):
        return self.get_tax().amount(net)
    
    def apply(self, net):
        return self.get_tax().apply(net)
    
    def reverse(self, gross):
        return self.get_tax().reverse(gross)
    
    def get_tax(self):
        raise RuntimeError('subclass must implement this')


class LinearTax(Tax):
    # TODO: PercentField?
    percent = models.DecimalField(max_digits=6, decimal_places=3)
    
    def get_tax(self):
        from . import LinearTax
        tax = LinearTax(self.name, self.percent)
        tax._unique_id = 'linear-pk-%d' % self.pk
        tax._model_instance = self
        return tax


class MultiTax(Tax):
    taxes = models.ManyToManyField(Tax, related_name='+')
    
    def get_tax(self):
        from . import MultiTax
        tax = MultiTax(list(self.taxes.all()), self.name)
        tax._unique_id = 'multi-pk-%d' % self.pk
        tax._model_instance = self
        return tax

